/****************************************************************************/
/*                                                                          */
/*  Author: David Lubicz, david.lubicz@univ-rennes1.fr                      */
/*  Last changes: xxth of October 2006                                      */
/*                                                                          */
/*  This program is distributed under the GNU General Public License (GPL)  */
/*  http://www.gnu.org/licenses/                                            */
/*                                                                          */
/****************************************************************************/
Attach("modeq.m");

/********************************************
/ definition of the lifting polynomial
/ as well as its partial derivatives
/ for elliptic curves
********************************************/
function phi(x, M1, evalvar, dec)
    local 
	SS, result, i, j, k;

    SS := Parent(x[1][1,1]);


    result := ZeroMatrix(SS, Nrows(M1), Ncols(M1));

    for i := 1 to Nrows(M1) do
	for j := 1 to Ncols(M1) do
	    for k := 1 to #evalvar do
		M1[i,j] := Evaluate(M1[i,j], evalvar[k],x[1][k,1]);
		M1[i,j] := Evaluate(M1[i,j], evalvar[k]+dec,x[2][k,1]);
	    end for;
	    if M1[i,j] eq 0 then 
		result[i,j] := 0;
	    else
		result[i,j] := Coefficients(M1[i,j])[1];
	    end if;
	end for;
    end for;

    return result;
end function;

//********************************************/
// miscellaneous matrix functions
//********************************************/
MValuation := function(x)
    local i,j,r;

r := Valuation(x[1,1]);
for i := 1 to Nrows(x) do
    for j := 1 to Ncols(x) do
	r := Min(r, Valuation(x[i,j]));
    end for;
end for;

return r;
end function;

//********************************************/
// idem for Precision
//********************************************/
MPrecision := function(x, m)
    local i,j;

for i := 1 to Nrows(x) do
    for j := 1 to Ncols(x) do
	ChangePrecision(~x[i,j],m);
    end for;
end for;

return x;
end function;

//********************************************/
// idem for Norm computation
//********************************************/
MNorm := function(x)
    local i,j;

for i := 1 to Nrows(x) do
    for j :=1 to Ncols(x) do
	x[i,j] := Norm(x[i,j]);
    end for;
end for;

return x;
end function;

//********************************************/
// Frobenius substitution
//********************************************/
function sigma(x,n)
    local result;

    result := x;

    for i := 1 to Nrows(x) do
	for j := 1 to Ncols(x) do
	    result[i,j] := GaloisImage(x[i,j],2*n);
	end for;
    end for;

    return result;
end function;

//********************************************/
// Frobenius substitution for finite fields
//********************************************/
function Fsigma(x,n)
    local result,c;

    result := x;
    c := Characteristic(Parent(x[1,1]));

    for i := 1 to Nrows(x) do
	for j := 1 to Ncols(x) do
	    result[i,j] := x[i,j]^c;
	end for;
    end for;

    return result;
end function;

//********************************************/
// Compute the isomorphism between finite field of the same cardinality
//********************************************/
function FiniteFieldConvert(x)
    local i;

    FF := Parent(x[1]);
    AFF := FiniteField(#FF);

    result := [];
    Embed(FF, AFF);

    for i := 1 to #x do
	result := Append(result, AFF!x[i]);
    end for;

    return result;
end function;

//********************************************/
// From a matrix x[i,j] compute the same matrix x[i,j] with coef in
// the quotient field of x[i,j]
//********************************************/
function MatrixQuoField(x)
    local i,j, FF, QF, Result;

    FF := Parent(x[1,1]);
    QF := FieldOfFractions(FF);

    Result := ZeroMatrix(QF, Nrows(x), Ncols(x));

    for i := 1 to Nrows(x) do
	for j :=1 to Ncols(x) do
	    Result[i,j] := QF!x[i,j];
	end for;
    end for;    

    return Result;
end function;


//********************************************/
// Lift a matrix over the padic
//********************************************/
function MatrixLift(x, precision)
    local i,j, FF, P, Zp, ZpX, Zpl, Result;

    FF := Parent(x[1,1]);
    ExtPol := DefiningPolynomial(FF);
    Zp := pAdicRing(Characteristic(FF),precision);
    ZpPolRing := PolynomialRing(Zp);
    Zpl := UnramifiedExtension(Zp, ZpPolRing!Coefficients(ExtPol));

    Result := ZeroMatrix(Zpl, Nrows(x), Ncols(x));

    for i := 1 to Nrows(x) do
	for j :=1 to Ncols(x) do
	    Result[i,j] := Zpl!Eltseq(x[i,j]);
	end for;
    end for;    

    return Result;
end function;

//********************************************/
// m is the precision to be reached nu is a parameter,
// a and b are the coefficients of the Artin-Schreier extension
//********************************************/
function ArtinSchreierRoot(a,b,m,nu)
    local k, ga, gb, ga1, gb1;

    if nu eq 1 then return a,b;
    end if;

    k := Floor(nu/2);	 
    ga, gb := $$(a,b,m, k);

    ga1 := sigma(ga, k)*ga;
    gb1 := sigma(ga, k)*gb+ sigma(gb,k);

    if (nu mod 2) eq 1 then 
	ga := ga1; gb := gb1;
	ga1 := sigma(ga,1) * a;
	gb1 := sigma(ga,1) * b + sigma(gb,1);
    end if;
    return ga1, gb1;

end function;

//********************************************/
// whereis x0 is the element to be lifted
// m is the precision to be reached
// phi is the lifting polynomial
// dphix and dphiy are the partial derivatives of this polynomial
//********************************************/
function Newtonlift(x0, m, n, M1, M2, M3, evalvar, dec)
    local padic,r,x1,x, y, t, result;

    padic := Parent(x0[1,1]);
    p := UniformizingElement(padic);
    x1 := sigma(x0, 1);

    r := MValuation(phi([x0, x1], M1, evalvar, dec)) - MValuation(phi([x0,
	x1], M3, evalvar, dec)^(-1)*phi([x0, x1], M1, evalvar, dec));

    if (2*r+1) lt MValuation(phi([x0, x1], M1, evalvar, dec)) then
	print "Not enough precision in the root...";
	return false;
    end if;

    if m le 2*r+1 then
	return x0;
    end if;

    w := Ceiling(m/2)+r;
    x := $$(x0, w, n, M1, M2, M3, evalvar, dec);
    y := sigma(x, 1);

    deltax := phi([x,y], M2, evalvar, dec); deltay := phi([x,y], M3, evalvar, dec);
    V := phi([x,y], M1, evalvar, dec);

    a,b := ArtinSchreierRoot(-deltay^(-1)*deltax, -deltay^(-1)*V/p^(w-r),
    w-r,n);
    id :=  ScalarMatrix(Nrows(a), padic!1);

    result := x + p^(w-r)*(id-a)^(-1) * b;
    return result;
end function;

//********************************************/
// From a matrix x[i,j] compute the same matrix x[i,j] with coef in
// the quotient field of x[i,j]
//********************************************/
function MatrixQuoField(x)
    local i,j, FF, QF, Result;

    FF := Parent(x[1,1]);

    QF := FieldOfFractions(FF);
    Result := ZeroMatrix(QF, Nrows(x), Ncols(x));

    for i := 1 to Nrows(x) do
	for j :=1 to Ncols(x) do
	    Result[i,j] := QF!x[i,j];
	end for;
    end for;    

    return Result;
end function;

//////////////////////////////////////
// Compute the reduction of a Polynomial
//////////////////////////////////////
DescendPolynome := function(P, Z, F)
    local Pf, R, p, i;
// Reduit un polynome a coefficients dans un anneau p-adique
// dans un corps fini dont il est le complete
liste := [];

Pf := PolynomialRing(F);
R, p := ResidueClassField(Z);
Embed(R,F);

for i in Coefficients(P) do	
    liste := Append(liste, F!p(i));	
end for;
return Pf!liste;
end function; //DescendPolynome

//////////////////////////////////////
// Square Root computation (if necessary in a field extension)
//////////////////////////////////////
MySqrt := function (A)
    local Z, Poly, i,j, P;

    Z := Parent(A[1]);
    Poly := PolynomialRing(Z);

    for i := 1 to #A do
	P := Poly.1^2 - A[i];
	if Roots(P) eq [] then
	    Z := ext<Z|P>;
	    Poly := PolynomialRing(Z);
	    A := [Z!j : j in A];
	end if;
	A[i]:= Sqrt(A[i]);
    end for;
    return A;
end function; // MySqrt

//////////////////////////////////////
// Lift a polynomial
//////////////////////////////////////
RelevePolynome := function(P, F, Z)
    // Releve un polynone a ceofficients dans le corps fini F dans
// dans son complete p-adique
liste := [];
Pl := PolynomialRing(Z);

for i in Coefficients(P) do	
    liste := Append(liste, Z!Eltseq(i));
end for;
return Pl!liste;
end function; //RelevePolynome

//////////////////////////////////////
// AGM Step
//////////////////////////////////////
AGMStep := function(List)
    local a, b;

a := 1/2*(List[1]+List[2]);
b := MySqrt([List[1]*List[2]])[1];

return [Parent(b)!a,b];
end function;


//////////////////////////////////////
// Reverse AGM step
//////////////////////////////////////
ReverseAGM := function(List)
    local x, a, b;

x := MySqrt([List[1]^2-List[2]^2])[1];


a := Parent(x)!List[1]+x;
b := Parent(x)!List[1]-x;

return [a,b];
end function;

//////////////////////////////////////
// Compute the 2-theta null points
//////////////////////////////////////
Coeff1 := function (q)
    local e0, e1, e2, a0, a1, a2, a3, List;
    
    Zpl := Parent(Coefficient(q, 0));
    
    //On calcule la factorisation de Delta dans Zpl
    Racinesq := [ i[1] : i in Roots(q)];
    
    for i in Racinesq do
	if i ne 0 then
	    if i ne 1 then
		List := MySqrt([i]);
		List := MySqrt([(1+List[1])/(1-List[1])]);
	    end if;
	end if;
    end for;
    
    a1 := List[1];
    
    List := [Parent(a1)!1, a1];
    
    return List;
end function;

//////////////////////////////////////
// Compute the 2-theta null points
//////////////////////////////////////
Coeff := function (q)
    local e0, e1, e2, a0, a1, a2, a3, List;
    
    Zpl := Parent(Coefficient(q, 0));
    
    //On calcule la factorisation de Delta dans Zpl
    Racinesq := [ i[1] : i in Roots(q)];
    
    e0 := Racinesq[1];
    e1 := Racinesq[2];
    e2 := Racinesq[3];
    a0 :=((e0-e1)*(e2-e1))^(-1);
    a1 :=((e1-e0)*(e2-e0))^(-1);
    
    List := [a0, a1];
    List := MySqrt(List);
    //    List := ReverseAGM(List);
    //    List := ReverseAGM(List);
    List := MySqrt(List);
    
    return List;
end function;

//////////////////////////////////////
// Compute the 4-theta null points
//////////////////////////////////////
ComputeThetaConstants := function(q)
    local List, a0, a1, a2;

List := Coeff1(q);
a0, a2 := Explode(List);
List := [1/2*(a0^2+a2^2)*a0*a2];
List := MySqrt(List);
List := MySqrt(List);
a1 := Explode(List);
return FiniteFieldConvert([a0, a1, a2]);
end function;

//////////////////////////////////////
// Recover12from6Theta
//////////////////////////////////////
Recover12from6Theta := function(List)
    local A,B,C,D, y0, y2, y4, y6, x0, x1, x2, x3, amb, amb1, i, j;

x0, x1, x2, x3 := Explode(List);
A := Sqrt(x0^2+x0*x3);
B := Sqrt(x0^2-x0*x3);
C := Sqrt(x2^2+x2*x3);
D := Sqrt(x2^2-x2*x3);
y0 := 1/2 * (A + B); 
y6 := 1/2 * (A - B);
y2 := 1/2 * (C + D);
y4 := 1/2 * (C - D);

amb :=[[y0, y6], [y6, y0], [-y0, -y6], [-y6, -y0]];
amb1 :=[[y2, y4], [y4, y2], [-y2, -y4], [-y4, -y2]];


for i in amb do
    for j in amb1 do
	if Valuation(x0*x2 - i[1]*y4 - i[2]*y2) ge 100 then
	    y0 := i[1];
	    y6 := i[2];
	    break;
	end if;
    end for;
end for;


return y0, y2, y4, y6;
end function;

//////////////////////////////////////
// Compute the 6-theta null points, the variable List contains the
//2-theta values
//////////////////////////////////////
ComputeThetaConstants10 := function(q, List, char)
    local F, Poly1, Poly, tmap,
    Maliste, REquations, A, FEquations,
    i, j,k ,r, I, Gb, NGb, P, b, v1, v2, MyList, flag, flag1, power, K;

F := Parent(List[1]);
char := Characteristic(F);
Poly1<y0,y1,y2,y3,y4,y5,y6,y7,y8,y9> := PolynomialRing(F, 10);
Poly<x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13,x14,x15,x16,x17,x18,x19> := PolynomialRing(F,20);
tmap := hom<Poly1 -> Poly | x0,x1,x2,x3,x4,x5,x6,x7,x8,x9>;

Maliste := [];

REquations := RiemannRelations(1, char, 1);
FEquations := FrobeniusSquareRelations(1,char,1);

for i := 1 to #REquations do
    ExtractRep(~REquations, ~r);
    r := Poly1!r;
    r := tmap(r);
    r := Evaluate(r, 1 ,List[1]);
    r := Evaluate(r, 6,List[2]);
    Maliste := Append(Maliste, r);
end for;

power := 2;

for i := 1 to 1 do
    ExtractRep(~FEquations, ~r);
    r := Poly!r;
    r := Evaluate(r, 11, List[1]^(char^power));
    r := Evaluate(r, 12, x1^(char^power));
    r := Evaluate(r, 13, x2^(char^power));
    r := Evaluate(r, 14, x3^(char^power));
    r := Evaluate(r, 15, x4^(char^power));
    r := Evaluate(r, 16, List[2]^(char^power));
    r := Evaluate(r, 1 ,List[1]);
    r := Evaluate(r, 6 ,List[2]);

    Maliste := Append(Maliste, r);
end for;

I := ideal<Poly| Maliste>;
SetVerbose("Groebner", 0); 
Gb := GroebnerBasis(I);
b, P := IsUnivariate(Gb[#Gb]);
v := [];


K := SplittingField(P);
if K eq F then
    for i in Roots(P) do
	if i[2] eq 25 then
	    v[4] := i[1];
	end if;
    end for;
    
    for j := 0 to 2 do
	NGb := [];
	for i in Gb do
	    Eq1 := Evaluate(i,5-j, v[4-j]);
	    if Eq1 ne 0 then
		NGb := Append(NGb, Eq1);
	    end if;
	end for;
	b,Eq1 := IsUnivariate(NGb[#NGb]);
	v[4-j-1] := Roots(Eq1)[1][1];
	Gb := NGb;
    end for;
    
    return F,[List[1], v[1], v[2], v[3], v[4], List[2]];
else
    return K, [];
end if;

end function;

//////////////////////////////////////
// Lift an element of a finite field to pAdics
//////////////////////////////////////
Lift2pAdic := function(x, precision)
    local Zp, PZp, P, ExtZp, lx;

Zp := pAdicRing(Characteristic(Parent(x)), precision);
PZp := PolynomialRing(Zp);

P := DefiningPolynomial(Parent(x));

ExtZp := UnramifiedExtension(Zp,P);
lx := ExtZp!Eltseq(x);

return lx;
end function;

//////////////////////////////////////
// Compute the lifting matrix from a list of equations
//////////////////////////////////////
ComputeLiftMatrix1 := function(ListEq, Listvar1, Listvar2)
    local M1, M2, M3, lenght, i,j;

length := #ListEq;

M1 := ZeroMatrix(Parent(ListEq[1]), length,1);
for i := 1 to length do
    M1[i,1] := ListEq[i];
end for;

M2 := ZeroMatrix(Parent(ListEq[1]), length, length);
for i := 1 to length do
    for j := 1 to length do
	M2[i,j] := Derivative(ListEq[i], Listvar1[j]);
    end for;
end for;

M3 := ZeroMatrix(Parent(ListEq[1]), length, length);
for i := 1 to length do
    for j := 1 to length do
	M3[i,j] := Derivative(ListEq[i], Listvar2[j]);
    end for;
end for;

return M1, M2, M3;
end function;

//////////////////////////////////////
// Compute the lifting matrix from a list of 2 equations
//////////////////////////////////////
ComputeLiftMatrix := function(ListEq)
    local M1, M2, M3;

M1 := Matrix(Parent(ListEq[1]), 2, 1, [ListEq[1], ListEq[2]]);
M2 := Matrix(Parent(ListEq[1]), 2, 2, [Derivative(ListEq[1], 1, 2),
    Derivative(ListEq[1], 1, 3), Derivative(ListEq[2], 1, 2),
    Derivative(ListEq[2], 1, 3)]);
M3 := Matrix(Parent(ListEq[1]), 2, 2, [Derivative(ListEq[1], 1, 8),
    Derivative(ListEq[1], 1, 9), Derivative(ListEq[2], 1, 8),
    Derivative(ListEq[2], 1, 9)]);

return M1, M2, M3;
end function;

//////////////////////////////////////
// Principal program
//////////////////////////////////////
n := 5;
char := 5;
genre := 2;
precision :=20;

Fd := FiniteField(char^n);
PFd<x> := PolynomialRing(Fd); 

//Here is the coefficient of the elliptic curve
//elliptic := [0, 1, Fd.1^78];
elliptic := [0, 1, Random(Fd)];

q := (x - elliptic[1])*(x - elliptic[2])*(x - elliptic[3]);

//these are the 4 theta null points a3 is not computed for reason of symmetry
a0, a1, a2 := Explode(ComputeThetaConstants(q));
//Computing the 6-theta null points starting with the same 2-theta
//null points given by Thomae
KK1, mylist := ComputeThetaConstants10(q, [a0, a2], char);
if mylist eq [] then
    elliptic := [KK1!x : x in elliptic];
    PFd<x> := PolynomialRing(KK1); 
    q := (x - elliptic[1])*(x - elliptic[2])*(x - elliptic[3]);
    a := KK1!a0;
    a2 := KK1!a2;
    KK1, mylist := ComputeThetaConstants10(q, [a0, a2], char);
end if;
v0, v1, v2, v3, v4, v5 := Explode(mylist);
KK := Parent(a0);
// We have to go over the field of definition of the 4-theta null points
elliptic := [KK!i : i in elliptic];
print elliptic;
PFd<z> := PolynomialRing(KK); 
q := (z - elliptic[1])*(z - elliptic[2])*(z - elliptic[3]);
E := EllipticCurve(q);
IsOrdinary(E);
print "The curve";
print E;
print "Base field";
print KK;
print "Polynome de definition";
print DefiningPolynomial(KK);

A2 := Lift2pAdic(a2, precision);
SS := Parent(A2);
KK := ResidueClassField(SS);

// Building Rieman and Frobenius equations
REquations := RiemannRelations(1, char, 1);
/* A := ReadFrobeniusRelationsP3FromDisk(1, 1, 1); */
/* FEquations := MakeInstanceOfSetOfFrobeniusRelationsP3(A, 1, 1, 1); */
FEquations := FrobeniusSquareRelations(1,char,1);

Poly1<z0,z1,z2,z3,z4,z5,z6,z7,z8,z9> := PolynomialRing(SS, 10);
Poly<x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,y0,y1,y2,y3,y4,y5,y6,y7,y8,y9> := PolynomialRing(SS,20);
tmap := hom<Poly1 -> Poly | x0,x1,x2,x3,x4,x5,x6,x7,x8,x9>;

Maliste := [];

// injecting the value of the lifted 4-theta null points in the Rieman equations
for i := 1 to 4 do
    ExtractRep(~REquations, ~r);
    r := Poly1!r;
    r := tmap(r);
    r := Evaluate(r, 1 , 1);
    r := Evaluate(r, 2, y1);
    r := Evaluate(r, 3, y2);
    r := Evaluate(r, 4 ,y3);
    r := Evaluate(r, 5 ,y4);
    r := Evaluate(r, 6 ,y5);
    r := Evaluate(r, 7 ,y6);
    r := Evaluate(r, 8 ,y7);
    r := Evaluate(r, 9 ,y8);
    r := Evaluate(r, 10,y9);

    Maliste := Append(Maliste, r);
end for;

// same thing for the Frobenius equations
for i := 1 to 1 do
    ExtractRep(~FEquations, ~r);
    r := Poly!r;
    r := Evaluate(r, 1, 1);
    r := Evaluate(r, 11, 1);
    Maliste := Append(Maliste, r);
end for;

M1, M2, M3 := ComputeLiftMatrix1(Maliste, [2,3,4,5,6], [12,13,14,15,16]);


//print "Theta constantes";
//print ComputeThetaConstants10(q, [a0, a2], char);

v1 := v1/v0;
v2 := v2/v0;
v3 := v3/v0;
v4 := v4/v0;
v5 := v5/v0;

Embed(Parent(v0),KK);

// These are the interesting 6-theta null points (non 0 and non 2
//torsion point
xx1 := SS!KK!Eltseq(v1);
xx2 := SS!KK!Eltseq(v2);
xx3 := SS!KK!Eltseq(v3);
xx4 := SS!KK!Eltseq(v4);
xx5 := SS!KK!Eltseq(v5);
deg := Degree(KK);

//print "5-adics";
//print SS;
//print DefiningPolynomial(SS);

sixtheta := MatrixQuoField(Matrix(Parent(xx1), 5, 1, [xx1, xx2,
    xx3, xx4, xx5]));
// for safety reason
//sixtheta := Matrix(Parent(xx1), 2, 1, [xx1, xx2]);
listvar := [2,3,4,5,6];
dec := 10;
power := 2;

/* print "*********************************************************";*/ 
/* print MValuation(phi([sixtheta, sigma(sixtheta,1)], M1, listvar, dec));*/ 
/* print Valuation(Determinant(phi([sixtheta, sigma(sixtheta,1)], M2, \*/ 
/*     listvar, dec)));*/ 
/* print Valuation(Determinant(phi([sixtheta, sigma(sixtheta,1)], M3, \*/ 
/*     listvar, dec))); */ 


Lsixtheta := Newtonlift(sixtheta, precision, deg, M1, M2, M3, listvar, \
    dec);
print "Lifted theta constants";
print Lsixtheta;
// This is the complete list of lifted six theta null points
Lsix := [ 1, Lsixtheta[1,1], Lsixtheta[2,1], Lsixtheta[3,1], \
    Lsixtheta[4,1], Lsixtheta[5,1]];

t1 := (1+2*Lsix[2]+2*Lsix[3]+2*Lsix[4]+2*Lsix[5]+Lsix[6])/(1+Lsix[6]);
t2 := Norm(t1);
print "============";
trace := t2 + 5^deg / t2;
print -trace;
print "############";
print Order(E)-#KK-1;
