/****************************************************************************/
/*                                                                          */
/*  Author: Robert Carls, David Lubicz                                      */
/*  emails: rcarls@gmx.de, david.lubicz@univ-rennes1.fr                     */
/*                                                                          */
/*  Last changes: 22th of April 2007                                        */
/*                                                                          */
/*  This program is distributed under the GNU General Public License (GPL)  */
/*  http://www.gnu.org/licenses/                                            */
/*                                                                          */
/****************************************************************************/

/* The following routines are used in order to generate the modular equations
   for every dimension and level. We provide procedures for storing the
   relations on disk and for retrieving them. Also TEX output is possible. */


///////////////////////////
// Some auxiliary functions
///////////////////////////


function AddMod(multiindex1,multiindex2,level)
    sum := <>;
    for i := 1 to #multiindex1 do
        Append(~sum,((multiindex1[i] + multiindex2[i]) mod level));
    end for;
    return sum;
end function;

function IsDivisible(multiindex,p)
    return &and[ ((number mod p) eq 0) : number in TupleToList(multiindex) ];
end function;

function IsEven(multiindex)
    return IsDivisible(multiindex,2);
end function;
    
function MinusMod(multiindex,level)
    inv := <>;
    for i := 1 to #multiindex do
        Append(~inv,((level-multiindex[i]) mod level));
    end for;
    return inv;
end function;

function DivideMod(multiindex,p)
    quotient := <>;
    for i := 1 to #multiindex do
        Append(~quotient,(multiindex[i] div p));
    end for;
    return quotient;
end function;

function HalfMod(multiindex)
    return DivideMod(multiindex,2);
end function;

function MultiplyMod(multiindex,n,level)
    product := <>;
    for i := 1 to #multiindex do
        Append(~product,((multiindex[i] * n) mod level));
    end for;
    return product;
end function;

function Normalize(multiindex,level)
    inverse := MinusMod(multiindex,level);
    for i := 1 to #multiindex do
        if (multiindex[i] gt inverse[i]) then
            return inverse;
        else
            if (multiindex[i] lt inverse[i]) then
                return multiindex;
            end if;
        end if;
    end for;
    return multiindex;
end function;

function MultiindexToString(multiindex)
    indexstring := "";
    for i := 1 to #multiindex do
        indexstring := indexstring cat IntegerToString(multiindex[i]);        
        if (i lt #multiindex) then
            indexstring := indexstring cat " ";
        end if;
    end for;
    return indexstring;
end function;

function MultiindexToTex(multiindex)
    indexstring := "";
    for i := 1 to #multiindex do
        indexstring := indexstring cat IntegerToString(multiindex[i]);        
        if (i lt #multiindex) then
            indexstring := indexstring cat ",";
        end if;
    end for;
    return indexstring;
end function;

function ConvertNumberseqToMultiindex(numberseq)
    multiindex := <>;
    for number in numberseq do
        Append(~multiindex,number); 
    end for;
    return multiindex;
end function;

function VarX(MPOL,multiindex,level)
    multiindex := Normalize(multiindex,level);
    index := 0;
    for i := 1 to #multiindex do
    index := index + multiindex[i] * level^(i-1); 
    end for;
    index := index + 1;
    return MPOL.index;
end function;

function VarY(MPOL,multiindex,level)
    multiindex := Normalize(multiindex,level);
    index := 0;
    for i := 1 to #multiindex do
    index := index + multiindex[i]*level^(i-1); 
    end for;
    index := index + 1 + ( Rank(MPOL) div 2);
    return MPOL.index;
end function;

function RepresentPreimagesXi1(fourtuple,level)
    list := {};
    a, b, c, d := Explode(TupleToList(fourtuple));
    combinations := { <a,b,c,d>, <a,c,b,d>, <a,d,c,b> };
    for comb in combinations do
        if (IsEven(AddMod(comb[1],comb[2],level)) and
            IsEven(AddMod(comb[3],comb[4],level))) then
            x := HalfMod(AddMod(comb[1],comb[2],level));                
            y := AddMod(comb[1],MinusMod(x,level),level);
            v := HalfMod(AddMod(comb[3],comb[4],level));                
            w := AddMod(comb[3],MinusMod(v,level),level);
            Include(~list,<x,y,v,w>);
        end if;
    end for;
    return list;
end function;

function MakePolynomial(formula)
    polynomial := {}; 
    I := MultisetToSet(formula);
    for term in I do
        m := Multiplicity(formula,term);
        Include(~polynomial,<m,term>);    
    end for;
    return polynomial;
end function;

function AddEquation(eq1,eq2)
    sum := {};
    monomials := { mon[2] : mon in (eq1 join eq2) };
    for elem in monomials do
        intersection := { mon : mon in eq1 | mon[2] eq elem } join
                        { mon : mon in eq2 | mon[2] eq elem };
        if not IsEmpty(intersection) then
            coeff := &+[ mon[1] : mon in intersection];
            if (coeff ne 0) then
	        Include(~sum,<coeff,elem>);
            end if;
        end if;
    end for;
    return sum;
end function;

function ReduceEquation(equation)
    reduction := {};
    coefficients := [ term[1] : term in equation ];
    content := Gcd(coefficients);
    for term in equation do
        Include(~reduction,<(term[1] div content),term[2]>);
    end for;
    return reduction;
end function;

function ReduceEquations(equationset);
    reducedset := {};
    for equation in equationset do
        Include(~reducedset,ReduceEquation(equation));
    end for;
    return reducedset;
end function;

function MinusEquation(equation)
    inverse := {};
    for term in equation do
        Include(~inverse,<-term[1],term[2]>);
    end for;
    return inverse;
end function;

function DifferenceEquation(eq1,eq2)
    return AddEquation(eq1,MinusEquation(eq2));
end function;

function MakeEquations(polynomials)
    equations := {};
    M := CartesianPower(polynomials,2);
    for pair in M do
        left := MakePolynomial(pair[1]); 
        right := MakePolynomial(pair[2]);
        f := DifferenceEquation(left,right);
        if not IsEmpty(f) then
            if (MinusEquation(f) notin equations) then
                Include(~equations,f);
            end if;
        end if;
    end for;
    return equations;
end function;


//////////////////////////////////////
// Generation of the Riemann Equations
//////////////////////////////////////


intrinsic ComputeRiemannRelations(g::RngIntElt,p::RngIntElt,exp2::RngIntElt) -> SetEnum
{ The Riemann relations are computed with respect to dimension g and
  level (2^exp2)*p. A sparse index representation is used.}

    assert ((p ge 2) and (g ge 1) and (exp2 ge 1));

    level := (2^exp2) * p;
    component := { i : i in [0..level-1] };

    if ((p gt 2) and (exp2 eq 1)) then
        halfcomponent := component;
    else
        halfcomponent := { i : i in component | (i mod 2) eq 0 };
    end if;

    halfrange := CartesianPower(halfcomponent,g);
    twocomponent := { 0, (level div 2) };
    twotorsion := CartesianPower(twocomponent,g);
    halffourtuples := CartesianPower(halfrange,4);

    S := {};
    for tuple in halffourtuples do
        preimages := RepresentPreimagesXi1(tuple,level);
        polynomials := {};
        for preim in preimages do
            formula := {* *};
            x, y, v, w := Explode(preim);
	    for t in twotorsion do
                for s in twotorsion do
	            monomial := {* Normalize(AddMod(x,t,level),level),
			           Normalize(AddMod(y,t,level),level),
			           Normalize(AddMod(v,s,level),level),
                                   Normalize(AddMod(w,s,level),level) *};
                    Include(~formula,monomial);
	        end for;
            end for;
            Include(~polynomials,formula);
        end for;
        if (#polynomials ge 2) then
	    equations := MakeEquations(polynomials);
            for elem in equations do
	        if (MinusEquation(elem) notin S) then
	            Include(~S,elem);
                end if;
            end for;
        end if;
    end for;

    S := ReduceEquations(S);
    return S;

end intrinsic;

function MakeInstanceOfRiemannRelation(indexset,g,p,exp2)

    assert ((p ge 2) and (g ge 1) and (exp2 ge 1));

    level := (2^exp2) * p;
    component := { i : i in [0..level-1] };
    range := CartesianPower(component,g);
    MPOL := PolynomialRing(Integers(),#range);

    polynomial := (MPOL!0); 
    for term in indexset do
        monomial := (MPOL!1);
        for multiindex in term[2] do
            monomial := monomial * VarX(MPOL,multiindex,level);
        end for;
        polynomial := polynomial + term[1] * monomial;
    end for;

    return polynomial;

end function;

intrinsic MakeInstanceOfSetOfRiemannRelations(polynomials::SetEnum,g::RngIntElt,p::RngIntElt,exp2::RngIntElt) -> SetEnum
{ An index set which represents a multivariate polynomial is translated
  to an element of a multivariate polynomial ring with (p*(2^exp2))^g
  elements. This is done for all index sets in the given set. }
    
    S := {};
    for poly in polynomials do
	Include(~S,MakeInstanceOfRiemannRelation(poly,g,p,exp2));
    end for;

    return S;

end intrinsic;

intrinsic WriteRiemannRelationsToDisk(polyset::SetEnum,g::RngIntElt,p::RngIntElt,exp2::RngIntElt)
{ The set of polynomials given by the variable polyset is stored in the
  working directory under the name 'riemann_g_p_exp2'.}

    filename := "riemann" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(p) cat "_" cat IntegerToString(exp2);
    file := Open(filename,"w");

    for polynomial in polyset do
        outputstring := "";
        monomiallist := SetToSequence(polynomial);
        for i := 1 to #monomiallist do
            monomial := monomiallist[i];
            outputstring := outputstring cat IntegerToString(monomial[1]);
            for multiindex in monomial[2] do
	        outputstring := outputstring cat " " cat MultiindexToString(multiindex);
            end for;
            if (i ne #polynomial) then
	        outputstring := outputstring cat " ";
            end if;
        end for;
        outputstring := outputstring cat "\n";
        Put(file,outputstring);        
    end for;

    Flush(file);

end intrinsic;

intrinsic ComputeAndWriteRiemannRelationsToDisk(g::RngIntElt,p::RngIntElt,exp2::RngIntElt)
{ This is a fully automated generation process for the Riemann equations of
  level (2^exp2)*p. The relations are stored as integer sequences.}

    ModEQ := ComputeRiemannRelations(g,p,exp2);
    WriteRiemannRelationsToDisk(ModEQ,g,p,exp2);

end intrinsic;

intrinsic ReadRiemannRelationsFromDisk(g::RngIntElt,p::RngIntElt,exp2::RngIntElt) -> SetEnum
{ This procedure reads the complete file and coverts every line into
  a polynomial. }

    S := {};

    filename := "riemann" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(p) cat "_" cat IntegerToString(exp2);
    file := Open(filename,"r");

    while true do
        line := Gets(file);
        if IsEof(line) then
            break;
        end if;
        polynomial := {};
        numberseq := StringToIntegerSequence(line);
        terms := Partition(numberseq,1+4*g);        
        for term in terms do
	    coeff := term[1];
            newterm := Remove(term,1);
            indices := Partition(newterm,g);
            monomial := {* *}; 
            for index in indices do
	        Include(~monomial,ConvertNumberseqToMultiindex(index));
            end for;
            Include(~polynomial,<coeff,monomial>);
	end for;
        Include(~S,polynomial);
    end while;

    return S;

end intrinsic;

intrinsic RiemannRelations(g::RngIntElt,p::RngIntElt,exp2::RngIntElt) -> SetEnum
{ This routine loads the Riemann relations from disk and converts them into
  multivariate polynomials. }

    S := ReadRiemannRelationsFromDisk(g,p,exp2);
    return MakeInstanceOfSetOfRiemannRelations(S,g,p,exp2);

end intrinsic;

intrinsic WriteRiemannRelationsToTexFile(polyset::SetEnum,g::RngIntElt,p::RngIntElt,exp2::RngIntElt)
{ Make a tex file with the Riemann equations. }

    filename := "riemann" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(p) cat "_" cat IntegerToString(exp2)
                cat ".tex";
    file := Open(filename,"w");

    header := "\\documentclass[10pt]{article}\n"
              cat "\\usepackage{a4wide}\n"
              cat "\\begin{document}\n"
              cat "\\small{\n"
              cat "\\begin{eqnarray*}\n";                             
    Put(file,header);

    for polynomial in polyset do
        outputstring := "& 0= &";
        monomiallist := SetToSequence(polynomial);
        for i := 1 to #monomiallist do
            monomial := monomiallist[i];
            if (Sign(monomial[1]) eq -1) then
                outputstring := outputstring cat "-"; 
            else
                if (i gt 1) then
                    outputstring := outputstring cat "+";
                end if;
            end if;
            if ((monomial[1] ne 1) and (monomial[1] ne -1)) then
                outputstring := outputstring cat IntegerToString(Sign(monomial[1])*monomial[1]);
            end if;
            strippedmonomial := MultisetToSet(monomial[2]);
            for multiindex in strippedmonomial do
	        m := Multiplicity(monomial[2],multiindex);
	        outputstring := outputstring cat "x_{" cat MultiindexToTex(multiindex) cat "}";
                if (m gt 1) then
                    outputstring := outputstring cat "^" cat IntegerToString(m);
                end if;
            end for;
            if (((i mod 4) eq 0) and (i ne #monomiallist)) then
                outputstring := outputstring cat "\\\\\n" cat "&&";
            end if;
        end for;
        outputstring := outputstring cat "\\\\\n";
        Put(file,outputstring);        
    end for;

    final := "\\end{eqnarray*}\n"
             cat "}\n"
	     cat "\\end{document}\n";
    Put(file,final);
    Flush(file);

end intrinsic;


//////////////////////////
// Frobenius relations p=3
//////////////////////////


function RepresentPreimagesXi3(threetuple,level)
    list := {};
    a, b, c := Explode(TupleToList(threetuple));
    combinations := { <a,b,c>, <b,c,a>, <c,b,a> };
    for comb in combinations do
        if (comb[3] eq MinusMod(AddMod(comb[1],comb[2],level),level)) then
            y := MinusMod(HalfMod(comb[1]),level);                
            z := AddMod(y,MinusMod(comb[2],level),level);
            Include(~list,<y,z>);
        end if;
    end for;
    return list;
end function;

intrinsic ComputeFrobeniusRelationsP3(g::RngIntElt,exp2::RngIntElt,exp3::RngIntElt) -> SetEnum
{ This routine computes the Frobenius Correspondence for level (2^exp*3^exp3)^g. }

    assert ((g ge 1) and (exp2 ge 1) and (exp3 ge 0));

    level := (2^exp2) * (3^exp3);
    component := { i : i in [0..level-1] };
    range := CartesianPower(component,g);
    halfcomponent := { i : i in component | (i mod 2) eq 0 };
    halfrange := CartesianPower(halfcomponent,g);
    twocomponent := { 0, (level div 2) };
    twotorsion := CartesianPower(twocomponent,g);
    halfthreetuples := CartesianPower(halfrange,3);

    S := {};
    for tuple in halfthreetuples do
        preimages := RepresentPreimagesXi3(tuple,level);
        polynomials := {};
        for preim in preimages do
            formula := {* *};
            y, z := Explode(preim);
	    for t in twotorsion do
	        monomial := < Normalize(AddMod(y,t,level),level),
		              Normalize(AddMod(z,t,level),level) >;
                Include(~formula,monomial);
            end for;
            Include(~polynomials,formula);
        end for;
        if (#polynomials ge 2) then
	    equations := MakeEquations(polynomials);
            for elem in equations do
	        if (MinusEquation(elem) notin S) then
	            Include(~S,elem);
                end if;
            end for;
        end if;
    end for;

    S := ReduceEquations(S);
    return S;

end intrinsic;

function MakeInstanceOfFrobeniusRelationP3(indexset,g,exp2,exp3)

    assert ((g ge 1) and (exp2 ge 1) and (exp3 ge 0));

    level := (2^exp2) * (3^exp3);
    component := { i : i in [0..level-1] };
    range := CartesianPower(component,g);
    MPOL := PolynomialRing(Integers(),2*#range);

    polynomial := (MPOL!0); 
    for term in indexset do
        monomial := VarY(MPOL,term[2][1],level) * VarX(MPOL,term[2][2],level);
        polynomial := polynomial + term[1] * monomial;
    end for;

    return polynomial;

end function;

intrinsic MakeInstanceOfSetOfFrobeniusRelationsP3(polynomials::SetEnum,g::RngIntElt,exp2::RngIntElt,exp3::RngIntElt) -> SetEnum
{ A set of Frobenius relations is converted in multivariate polynomials }
    
    S := {};
    for poly in polynomials do
	Include(~S,MakeInstanceOfFrobeniusRelationP3(poly,g,exp2,exp3));
    end for;
    return S;

end intrinsic;

intrinsic WriteFrobeniusRelationsP3ToDisk(polyset::SetEnum,g::RngIntElt,exp2::RngIntElt,exp3::RngIntElt)
{ The set of polynomials given by the variable polyset is stored in the
  working directory under the name 'frobenius_g_exp2_exp3'.}

    filename := "frobenius3" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(exp2) cat "_" cat IntegerToString(exp3);
    file := Open(filename,"w");

    for polynomial in polyset do
        outputstring := "";
        monomiallist := SetToSequence(polynomial);
        for i := 1 to #monomiallist do
            monomial := monomiallist[i];
            outputstring := outputstring cat IntegerToString(monomial[1])
                            cat " " cat MultiindexToString(monomial[2][1])
                            cat " " cat MultiindexToString(monomial[2][2]);
            if (i ne #monomiallist) then
	        outputstring := outputstring cat " ";
            end if;
        end for;
        outputstring := outputstring cat "\n";
        Put(file,outputstring);        
    end for;

    Flush(file);

end intrinsic;

intrinsic ComputeAndWriteFrobeniusRelationsP3ToDisk(g::RngIntElt,exp2::RngIntElt,exp3::RngIntElt)
{ This is a fully automated generation process for the Frobenius relations of
  level (2^exp2)*(3^exp3). The relations are stored as integer sequences.}

    ModEQ := ComputeFrobeniusRelationsP3(g,exp2,exp3);
    WriteFrobeniusRelationsP3ToDisk(ModEQ,g,exp2,exp3);

end intrinsic;

intrinsic ReadFrobeniusRelationsP3FromDisk(g::RngIntElt,exp2::RngIntElt,exp3::RngIntElt) -> SetEnum
{ This procedure reads the complete file and coverts every line into
  a polynomial. }

    S := {};

    filename := "frobenius3" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(exp2) cat "_" cat IntegerToString(exp3);
    file := Open(filename,"r");

    while true do
        line := Gets(file);
        if IsEof(line) then
            break;
        end if;
        polynomial := {};
        numberseq := StringToIntegerSequence(line);
        terms := Partition(numberseq,1+2*g);
        for term in terms do
	    coeff := term[1];
            newterm := Remove(term,1);
            indices := Partition(newterm,g);
            monomial := <>; 
            for index in indices do
	        Append(~monomial,ConvertNumberseqToMultiindex(index));
            end for;
            Include(~polynomial,<coeff,monomial>);
	end for;
        Include(~S,polynomial);
    end while;

    return S;

end intrinsic;

intrinsic WriteFrobeniusRelationsP3ToTexFile(polyset::SetEnum,g::RngIntElt,exp2::RngIntElt,exp3::RngIntElt)
{ Make a tex file with the Frobenius P3 equations. }

    filename := "frobenius3" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(exp2) cat "_" cat IntegerToString(exp3)
                cat ".tex";
    file := Open(filename,"w");

    header := "\\documentclass[10pt]{article}\n"
              cat "\\usepackage{a4wide}\n"
              cat "\\begin{document}\n"
              cat "\\small{\n"
              cat "\\begin{eqnarray*}\n";                             
    Put(file,header);

    for polynomial in polyset do
        outputstring := "& 0= &";
        monomiallist := SetToSequence(polynomial);
        for i := 1 to #monomiallist do
            monomial := monomiallist[i];
            if (Sign(monomial[1]) eq -1) then
                outputstring := outputstring cat "-"; 
            else
                if (i gt 1) then
                    outputstring := outputstring cat "+";
                end if;
            end if;
            if ((monomial[1] ne 1) and (monomial[1] ne -1)) then
                outputstring := outputstring cat IntegerToString(Sign(monomial[1])*monomial[1]);
            end if;
	    outputstring := outputstring cat "x^{\\sigma}_{"
                            cat MultiindexToTex(monomial[2][1]) cat "}"
		            cat "x_{" cat MultiindexToTex(monomial[2][2]) cat "}";
            if (((i mod 4) eq 0) and (i ne #monomiallist)) then
                outputstring := outputstring cat "\\\\\n" cat "&&";
            end if;
        end for;
        outputstring := outputstring cat "\\\\\n";
        Put(file,outputstring);        
    end for;

    final := "\\end{eqnarray*}\n"
             cat "}\n"
	     cat "\\end{document}\n";
    Put(file,final);
    Flush(file);

end intrinsic;

intrinsic FrobeniusRelationsP3(g::RngIntElt,exp2::RngIntElt,exp3::RngIntElt) -> SetEnum
{ This routine loads the Frobenius relations from disk and converts them into
  multivariate polynomials. }

    S := ReadFrobeniusRelationsP3FromDisk(g,exp2,exp3);
    return MakeInstanceOfSetOfFrobeniusRelationsP3(S,g,exp2,exp3);

end intrinsic;


///////////////////////////////////////////
// Computation of the Frobenius^2 relations
///////////////////////////////////////////


function RepresentPreimagesXip(twotuple,p,level)
    list := {};
    a, b := Explode(TupleToList(twotuple));
    combinations := { <a,b>, <b,a> };

/*     half(a+b);
       divide-by-p(half(a-b)); */

    for comb in combinations do
        if IsEven(AddMod(comb[1],comb[2],level)) then
            v := HalfMod(AddMod(comb[1],comb[2],level));
            z := AddMod(comb[1],MinusMod(v,level),level);
            if IsDivisible(z,p) then
                w := DivideMod(z,p);                
                Include(~list,<v,w>);
            end if;
	end if;
    end for;
    return list;
end function;

intrinsic ComputeFrobeniusSquareRelations(g::RngIntElt,p::RngIntElt,exp2::RngIntElt) -> SetEnum
{ This routine computes the Frobenius^2 Correspondence for level (2^exp2*p)^g. }

assert ((g ge 1) and (exp2 ge 1) and (p ge 2));

level := (2^exp2) * p;
component := { i : i in [0..level-1] };
range := CartesianPower(component,g);

if (exp2 eq 1) then
    halfcomponent := component;
else
    halfcomponent := { i : i in component | (i mod 2) eq 0 };
end if;

halfrange := CartesianPower(halfcomponent,g);
twopecomponent := { i : i in component | ((i mod (level div (2*p))) eq 0) };
twopetorsion := CartesianPower(twopecomponent,g);
halftwotuples := CartesianPower(halfrange,2);
twocomponent := { 0, (level div 2) };
twotorsion := CartesianPower(twocomponent,g);

S := {};
for tuple in halftwotuples do
    polynomials1 := {};
    a, b := Explode(TupleToList(tuple));
    if IsEven(AddMod(a,b,level)) then
        x := HalfMod(AddMod(a,b,level));                
        y := AddMod(a,MinusMod(x,level),level);
        formula := {* *};
	for t in twotorsion do
	    monomial := < Normalize(AddMod(x,t,level),level),
			  Normalize(AddMod(y,t,level),level) >;
            Include(~formula,monomial);
        end for;
        Include(~polynomials1,MakePolynomial(formula));
    end if;
    preimages := RepresentPreimagesXip(tuple,p,level);
    polynomials2 := {};
    for preim in preimages do
        formula := {* *};
        x, y := Explode(preim);
	for t in twopetorsion do
	    monomial := < Normalize(AddMod(x,MultiplyMod(t,p,level),level),level), Normalize(AddMod(y,t,level),level) >;
            Include(~formula,monomial);
        end for;
        Include(~polynomials2,MakePolynomial(formula));
    end for;
    if ((#polynomials1 ge 1) and (#polynomials2 ge 1)) then
	equations := { <pol[1],pol[2]> : pol in CartesianProduct(polynomials1,polynomials2) };
        for elem in equations do
	    if (<MinusEquation(elem[1]),MinusEquation(elem[2])> notin S) then
	        Include(~S,elem);
            end if;
        end for;
    end if;
end for;

return S;

end intrinsic;

function MakeInstanceOfPairOfFrobeniusSquareRelations(pair1,pair2,g,p,exp2)

    assert ((g ge 1) and (exp2 ge 1) and (p ge 2));

    level := (2^exp2) * p;
    component := { i : i in [0..level-1] };
    range := CartesianPower(component,g);
    indexmax := 2*#range;
    MPOL := PolynomialRing(Integers(),indexmax);

    polynomial := (MPOL!0);
    for term in pair1[1] do
        monomial := VarX(MPOL,term[2][1],level) * VarX(MPOL,term[2][2],level);
        polynomial := polynomial + term[1] * monomial;
    end for;
    lefthandside := polynomial;
    polynomial := (MPOL!0);
    for term in pair2[2] do
        monomial := VarX(MPOL,term[2][1],level) * VarY(MPOL,term[2][2],level);
        polynomial := polynomial + term[1] * monomial;
    end for;
    lefthandside := lefthandside * polynomial;
    polynomial := (MPOL!0);
    for term in pair1[2] do
        monomial := VarX(MPOL,term[2][1],level) * VarY(MPOL,term[2][2],level);
        polynomial := polynomial + term[1] * monomial;
    end for;
    righthandside := polynomial;
    polynomial := (MPOL!0);    
    for term in pair2[1] do
        monomial := VarX(MPOL,term[2][1],level) * VarX(MPOL,term[2][2],level);
        polynomial := polynomial + term[1] * monomial;
    end for;
    righthandside := righthandside * polynomial;
    return (lefthandside - righthandside);

end function;

intrinsic MakeInstanceOfSetOfFrobeniusSquareRelations(denom::Tup,polynomials::SetEnum,g::RngIntElt,p::RngIntElt,exp2::RngIntElt) -> SetEnum
{ A set of Frobenius relations is converted in multivariate polynomials. }

    S := {};
    for poly in polynomials do
	Include(~S,MakeInstanceOfPairOfFrobeniusSquareRelations(poly,denom,g,p,exp2));
    end for;
    return S;

end intrinsic;

intrinsic WriteFrobeniusSquareRelationsToDisk(equationset::SetEnum,g::RngIntElt,p::RngIntElt,exp2::RngIntElt)
{ The set of polynomials given by the variable polyset is stored in the
  working directory under the name 'frobeniussq_g_p_exp2'. One equation is written in
  two lines. }

    filename := "frobeniussq" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(p) cat "_" cat IntegerToString(exp2);
    file := Open(filename,"w");

    for equation in equationset do
        for monomialset in TupleToList(equation) do
            outputstring := "";
            monomiallist := SetToSequence(monomialset);
            for i := 1 to #monomiallist do
                monomial := monomiallist[i];
                outputstring := outputstring cat IntegerToString(monomial[1])
                                cat " " cat MultiindexToString(monomial[2][1])
                                cat " " cat MultiindexToString(monomial[2][2]);
                if (i ne #monomiallist) then
	            outputstring := outputstring cat " ";
                end if;
            end for;
            outputstring := outputstring cat "\n";
            Put(file,outputstring);
        end for;
    end for;

    Flush(file);

end intrinsic;

intrinsic ComputeAndWriteFrobeniusSquareRelationsToDisk(g::RngIntElt,p::RngIntElt,exp2::RngIntElt)
{ This is a fully automated generation process for the p-Frobenius relations of
  level (2^exp2)*p. The relations are stored as integer sequences.}

    ModEQ := ComputeFrobeniusSquareRelations(g,p,exp2);
    WriteFrobeniusSquareRelationsToDisk(ModEQ,g,p,exp2);

end intrinsic;

intrinsic ReadFrobeniusSquareRelationsFromDisk(g::RngIntElt,p::RngIntElt,exp2::RngIntElt) -> SetEnum
{ This procedure reads the complete file and coverts every two lines into
  a polynomial. }

    S := {};

    filename := "frobeniussq" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(p) cat "_" cat IntegerToString(exp2);
    file := Open(filename,"r");

    while true do
        line1 := Gets(file);
        if IsEof(line1) then
            break;
        end if;
        line2 := Gets(file);
        numberseq1 := StringToIntegerSequence(line1);
        numberseq2 := StringToIntegerSequence(line2);
        polynomials := [ Partition(numberseq1,1+2*g), Partition(numberseq2,1+2*g) ];
        equation := <>;
        for polynomial in polynomials do
	    terms := {};
            for term in polynomial do
	        coeff := term[1];
                newterm := Remove(term,1);
                indices := Partition(newterm,g);
                monomial := <>; 
                for index in indices do
	            Append(~monomial,ConvertNumberseqToMultiindex(index));
                end for;
                Include(~terms,<coeff,monomial>);
	    end for;
            Append(~equation,terms);
        end for;
        Include(~S,equation);
    end while;

    return S;

end intrinsic;

intrinsic FrobeniusSquareRelations(g::RngIntElt,p::RngIntElt,exp2::RngIntElt) -> SetEnum
{ This routine loads the Frobenius square relations from disk and converts them into
  multivariate polynomials. }

    S := ReadFrobeniusSquareRelationsFromDisk(g,p,exp2);
    denom := Random(S);
    Exclude(~S,denom);

    return MakeInstanceOfSetOfFrobeniusSquareRelations(denom,S,g,p,exp2);

end intrinsic;

intrinsic WriteFrobeniusSquareRelationsToTexFile(equationset::SetEnum,g::RngIntElt,p::RngIntElt,exp2::RngIntElt)
{ Make a tex file with the Frobenius square equations. }

    filename := "frobeniussq" cat "_" cat IntegerToString(g) cat "_"
                cat IntegerToString(p) cat "_" cat IntegerToString(exp2)
                cat ".tex";
    file := Open(filename,"w");

    header := "\\documentclass[10pt]{article}\n"
              cat "\\usepackage{a4wide}\n"
              cat "\\begin{document}\n"
              cat "\\small{\n"
              cat "\\begin{eqnarray*}\n";                             
    Put(file,header);

    for equation in equationset do
        outputstring :="";
        monomiallist := SetToSequence(equation[1]);
        for i := 1 to #monomiallist do
            monomial := monomiallist[i];
            if (Sign(monomial[1]) eq -1) then
                outputstring := outputstring cat "-"; 
            else
                if (i gt 1) then
                    outputstring := outputstring cat "+";
                end if;
            end if;
            if ((monomial[1] ne 1) and (monomial[1] ne -1)) then
                outputstring := outputstring cat IntegerToString(Sign(monomial[1])*monomial[1]);
            end if;
	    outputstring := outputstring cat "x_{"
                            cat MultiindexToTex(monomial[2][1]) cat "}"
		            cat "x_{" cat MultiindexToTex(monomial[2][2]) cat "}";
            if (((i mod 4) eq 0) and (i ne #monomiallist)) then
                outputstring := outputstring cat "\\\\\n" cat "&&";
            end if;
        end for;
        outputstring := outputstring cat " & = & \\omega ("; 
        monomiallist := SetToSequence(equation[2]);
        for i := 1 to #monomiallist do
            monomial := monomiallist[i];
            if (Sign(monomial[1]) eq -1) then
                outputstring := outputstring cat "-"; 
            else
                if (i gt 1) then
                    outputstring := outputstring cat "+";
                end if;
            end if;
            if ((monomial[1] ne 1) and (monomial[1] ne -1)) then
                outputstring := outputstring cat IntegerToString(Sign(monomial[1])*monomial[1]);
            end if;
	    outputstring := outputstring cat "x_{"
                            cat MultiindexToTex(monomial[2][1]) cat "}"
		            cat "x^{\\sigma^2}_{" cat MultiindexToTex(monomial[2][2]) cat "}";
            if (((i mod 4) eq 0) and (i ne #monomiallist)) then
                outputstring := outputstring cat "\\\\\n" cat "&&";
            end if;
        end for;
        outputstring := outputstring cat ") \\\\\n";
        Put(file,outputstring);        
    end for;

    final := "\\end{eqnarray*}\n"
             cat "}\n"
	     cat "\\end{document}\n";
    Put(file,final);
    Flush(file);

end intrinsic;

